#ifndef VECTORREF_H
#define VECTORREF_H

#include <string.h>
#include "Alloc.h"


template <class T, int N>
class Vector
{
private: 
 
  T* dataStart;
  T* dataFinish;

  unsigned int *repNr;
  
  static Allocator<T,N>             allocator_data;
  static Allocator<unsigned int, 1> allocator_repNr;
  
public:
  Inline Vector();
  Inline Vector(const Vector<T, N>& v);
  Inline ~Vector();

  Inline T&            operator[](int n);
  
  Inline Vector<T, N>& operator=(const Vector<T, N>& v);
  Inline Vector<T, N>  operator+(const Vector<T, N>& v) const;
  Inline Vector<T, N>  operator-(const Vector<T, N>& v) const;
  Inline T             operator*(const Vector<T, N>& v) const;
  Inline Vector<T,N>   operator*(T t) const;
};


template <class T, int N>
Inline Vector<T, N>::Vector(void) 
      : dataStart(allocator_data.allocate()),
	dataFinish(dataStart+N),
	repNr(allocator_repNr.allocate())
{
  *repNr=1;
}


template <class T, int N>
Inline Vector<T, N>::Vector(const Vector<T, N>& v)
      : dataStart(v.dataStart),
	dataFinish(v.dataFinish),
	repNr(v.repNr)
{
  (*repNr)++;
}


template <class T, int N>
Inline Vector<T, N>::~Vector()
{
  if (!(--(*repNr)))
    {
      allocator_repNr.deallocate(repNr);
      allocator_data.deallocate(dataStart);
    }
}


template <class T, int N>
Inline Vector<T, N>& Vector<T, N>::operator=(const Vector<T, N>& v)
{
  if (!(--(*repNr)))
    {
      allocator_repNr.deallocate(repNr);
      allocator_data.deallocate(dataStart);
    }
  
  (*v.repNr)++;
  repNr=v.repNr;

  dataStart=v.dataStart;
  dataFinish  =v.dataFinish;  

  return *this;
}
  

#ifdef NOSPECIAL

template <class T, int N>
Inline T& Vector<T, N>::operator[](int n)
{
  if ((*repNr) == 1)
    return dataStart[n];
  else
    {
      (*repNr)--;
      repNr=allocator_repNr.allocate();
      *repNr=1;

      T* temp=allocator_data.allocate();
      memcpy( (void*) temp, (void*) dataStart,
	     size_t(N*sizeof(T)) );
      
      dataStart=temp;
      dataFinish  =dataStart+N;
      
      return dataStart[n];
    }
}

#else

template <class T, int N>
Inline T& Vector<T, N>::operator[](int n)
{
  if ((*repNr) == 1)
    return dataStart[n];
  else
    {
      (*repNr)--;
      repNr=allocator_repNr.allocate();
      *repNr=1;

      T *src =dataStart;
      T *dest=dataStart=allocator_data.allocate();
      
      *(dest++)=*(src++);
      *(dest++)=*(src++);
      *dest    =*src;      
      
      return dataStart[n];
    }
}

#endif

#ifdef NOSPECIAL
template <class T, int N>
Inline Vector<T, N> Vector<T, N>::operator+(const Vector<T, N>& v) const
{
  Vector<T, N> tmp;

  T* dest=tmp.dataStart;
  T* sum1=v.dataStart;  
  T* sum2=dataStart;

  do
    *(dest++)=*(sum1++) + *(sum2++);
  while (sum2 != dataFinish);
    
  return tmp;
}

#else

Inline Vector<double, 3>
Vector<double, 3>::operator+(const Vector<double, 3>& v) const
{
  Vector<double, 3> tmp;

  double* dest=tmp.dataStart;
  double* sum1=v.dataStart;
  double* sum2=dataStart;

  *(dest++)=*(sum1++) + *(sum2++);
  *(dest++)=*(sum1++) + *(sum2++);
  *dest    =*sum1     + *sum2;
    
  return tmp;
}
#endif


#ifdef NOSPECIAL
template <class T, int N>
Inline Vector<T, N> Vector<T, N>::operator-(const Vector<T, N>& v) const
{
  Vector<T, N> tmp;

  T* dest=tmp.dataStart;
  T* sum1=v.dataStart;  
  T* sum2=dataStart;

  do
    *(dest++)=*(sum2++) - *(sum1++);
  while (sum2 != dataFinish);
    
  return tmp;
}

#else

Inline Vector<double, 3>
Vector<double, 3>::operator-(const Vector<double, 3>& v) const
{
  Vector<double, 3> tmp;

  double* dest=tmp.dataStart;
  double* sum1=v.dataStart;  
  double* sum2=dataStart;

  *(dest++)=*(sum2++) - *(sum1++);
  *(dest++)=*(sum2++) - *(sum1++);
  *dest    =*sum2     - *sum1;
    
  return tmp;
}
#endif


#ifdef NOSPECIAL
template <class T, int N>
Inline T Vector<T, N>::operator*(const Vector<T, N>& v) const
{
  T result(0);

  T* sum1=v.dataStart;  
  T* sum2=dataStart;

  do
    result+=*(sum1++) * *(sum2++);
  while (sum2 != dataFinish);
    
  return result;
}

#else

Inline double
Vector<double, 3>::operator*(const Vector<double, 3>& v) const
{
  double result;

  double* sum1=v.dataStart;  
  double* sum2=dataStart;

  result =*(sum1++) * *(sum2++);
  result+=*(sum1++) * *(sum2++);
  result+=*sum1 * *sum2;

  return result;
}
#endif

#ifdef NOSPECIAL
template <class T, int N>
Inline Vector<T, N> Vector<T, N>::operator*(T t) const
{
  Vector<T, N> tv(*this);

  T* ptr(tv.dataStart);
  do
    *(ptr++) *= t;
  while (ptr != tv.dataFinish);
  
  return tv;
}

#else

Inline Vector<double, 3> Vector<double, 3>::operator*(double t) const
{
  Vector<double,3> tv(*this);

  double* ptr(tv.dataStart);
  *(ptr++) *= t;
  *(ptr++) *= t;
  *(ptr) *= t;
  return tv;
}

#endif

#endif


// Local Variables:
// mode: c++
// End:
